
#pragma once
//#include "../myincludes.hh"
#include <dune/phasefield/base_potentials.hh>
#include <cmath>
#include <limits>
#include <dune/common/parametertreeparser.hh>
#include <iostream>


//______________________________________________________________________________________________________

template<typename Number>
class Mobility_Const
{
public:
  const Number mobility;
  Mobility_Const(const Dune::ParameterTree & param) :
    mobility(param.get("Mobility",(Number)0.1))
  {
  }

  Number eval( const Number& pf) const
  {
    return mobility;
  }

  Number eval( const Number& pf1, const Number& pf2, const Number& pf3) const
  {
    return mobility;
  }
};

template<typename Number>
class Mobility_Degen
{
public:
  const Number mobility;
  const Number mobility_reg;
  Mobility_Degen(const Dune::ParameterTree & param) :
    mobility(param.get("Mobility",(Number)0.1)),
    mobility_reg(param.get("MobilityMin",(Number)0.0))
  {
  }

  Number eval( const Number& pf) const
  {
    return std::max(mobility*pf*(1-pf),mobility_reg);
  }

  Number eval( const Number& pf1, const Number& pf2, const Number& pf3) const
  {
    std::cout << "Mobility_Degen for three phases not properly implemented yet" << std::endl;
    return mobility;
  }
};

//______________________________________________________________________________________________________

template<typename Number>
class Reaction_Quadratic_Limited
{
public:
  const Number reactionRate;
  const Number reactionLimiter;
  Reaction_Quadratic_Limited(const Dune::ParameterTree & param) :
    reactionRate(param.get("ReactionRate",(Number)0.1)),
    reactionLimiter(param.get("ReactionLimiter",(Number)100))
  {
  }

  Number eval( const Number& u) const
  {
    if(u<0)
      return reactionRate*(-1.0+reactionLimiter*u);
    else
      return reactionRate*(u*u-1.0);
  }
};

template<typename Number>
class Reaction_Calcite_Precipitation // Calcite precipitation CO3^2- + Ca^2+ -> CaCO3
                                     // unit of concentration in mol[CO3] /l (equilibrium at around 0.14*10^-3 mol/l)
                                     // we only know the reaction is happening faster than the timescale of seconds
                                     // As we assume the concentration of Ca^2+ to be big (around 0.16 mol/l)
                                     // we get a linear reaction kinetic
{
public:
  const Number reactionRate;
  const Number equilibriumConcentration;
  const Number bulkRate;
  Reaction_Calcite_Precipitation(const Dune::ParameterTree & param) :
    reactionRate(param.get("ReactionRate",(Number)10)),
    equilibriumConcentration(param.get("uEquilibrium",(Number)0.00014)),
    bulkRate(param.get("BulkRate",(Number)0.00001))
  {
  }

  Number eval( const Number& u) const
  {
    return reactionRate*(u-equilibriumConcentration);
  }

  Number eval_bulkreaction() const
  {
    return bulkRate;
  }
};

template<typename Number>
class Reaction_None
{
public:
  Reaction_None(const Dune::ParameterTree & param)
  {
  }

  Number eval( const Number& u) const
  {
    return 0;
  }

  Number eval_bulkreaction() const { return 0;}
};

//______________________________________________________________________________________________________

template<typename Number>
class VelDissipation_Quadratic
{
public:
  const Number velDissipationRate;
  VelDissipation_Quadratic(const Dune::ParameterTree & param) :
    velDissipationRate(param.get("VelDissipationRate",(Number)10))
  {
  }

  Number eval( const Number& pf_f) const
  {
    return velDissipationRate*(1-pf_f)*(1-pf_f);
  }
};

//______________________________________________________________________________________________________


template<typename Number>
class Params_Physical
{
public:
  const Number D;
  const Number uAst;
  const Number mu;
  const Number rho_1;
  const Number rho_2;
  const Number rho_3;
  const Number curvatureAlpha;

  Params_Physical(const Dune::ParameterTree& ptree, const Number eps) :
    D(ptree.get("D",(Number)0.1)),
    uAst(ptree.get("uAst",(Number)2)),
    mu(ptree.get("mu",(Number)0.01)),
    rho_1(ptree.get("rho1",(Number)1)),
    rho_2(ptree.get("rho2",(Number)1)),
    rho_3(ptree.get("rho1",(Number)1)),         //WE ASSUME RHO1 = RHO3 !!!
    curvatureAlpha(ptree.get("CurvatureAlpha", eps))
  {
  }

  Number rho_f (const Number& pf1, const Number& pf2) const
  {
    return pf1*rho_1+pf2*rho_2;
  }

};

//______________________________________________________________________________________________________
class Params_Time
{
public:
  double t;
  double dt;
  Params_Time(double t_, double dt_) :
    t(t_),dt(dt_)
  {
  }
};

//______________________________________________________________________________________________________

class Params_Initial
{
public:
  const Dune::ParameterTree& ptree;
  Params_Initial(const Dune::ParameterTree& tree) :
    ptree(tree)
  {
  }
};

//______________________________________________________________________________________________________

template <typename Number, typename Mobility>
class Params_base
{
public:
  Number eps;

  Mobility mobility;
  Params_Physical<Number> phys;
  Params_Time time;

  Params_Initial initial;


  Params_base(const Dune::ParameterTree& ptree) :
    eps(ptree.get("Phasefield.eps",(Number)0.1)),
    mobility(ptree.sub("Phasefield")),
    phys(ptree.sub("Physics"),eps),
    time(0,0),
    initial(ptree.sub("Initial"))
  {
  }
};


template <typename Number, typename DW, typename Mobility>
class Params_ff : public Params_base<Number, Mobility>
{
public:

  DW dw;
  Number SurfaceTensionScaling;

  Params_ff(const Dune::ParameterTree& ptree) :
    Params_base<Number, Mobility>(ptree),
    dw(ptree.sub("Phasefield")),
    SurfaceTensionScaling(ptree.get("Physics.SurfaceTensionScaling",(Number)1))
  {
  }
};


template <typename Number, typename TW, typename Mobility, typename Reaction, typename VelDissipation>
class Params_ffs_r : public Params_base<Number, Mobility>
{
public:
  Number delta;

  TW tw;
  Reaction reaction;
  VelDissipation vDis;

  Number SurfaceTensionScaling;


  Params_ffs_r(const Dune::ParameterTree& ptree) :
    Params_base<Number, Mobility>(ptree),
    delta(ptree.get("Phasefield.delta",this->eps)),
    tw(ptree.sub("Phasefield")),
    reaction(ptree.sub("Physics")),
    vDis(ptree.sub("Physics")),
    SurfaceTensionScaling(ptree.get("Physics.SurfaceTensionScaling",(Number)1))
  {
  }
};


template <typename Number, typename DW, typename Mobility, typename Reaction, typename VelDissipation>
class Params_fs_r : public Params_base<Number, Mobility>
{
public:
  Number delta;

  DW dw;
  Reaction reaction;
  VelDissipation vDis;

  Params_fs_r(const Dune::ParameterTree& ptree) :
    Params_base<Number, Mobility>(ptree),
    delta(ptree.get("Phasefield.delta",this->eps)),
    dw(ptree.sub("Phasefield")),
    reaction(ptree.sub("Physics")),
    vDis(ptree.sub("Physics"))
  {
  }
};

template <typename Number, typename DW, typename Mobility, typename Reaction, typename VelDissipation>
class Params_summerschool : public Params_fs_r<Number, DW, Mobility, Reaction, VelDissipation>
{
public:
  Number CellHeight;

  Params_summerschool(const Dune::ParameterTree& ptree) :
    Params_fs_r<Number, DW, Mobility, Reaction, VelDissipation>(ptree),
    CellHeight(ptree.get("domain.CellHeight",0.1))
  {
  }
};
//
// template<typename Number>
// class Problem
// {
//   Number t;
//
//
// public:
//   typedef Number value_type;
//   Number eps, delta;
//
//   Number dt;
//
//   Number u_regularization;
//
//
//   Number M;
//   Number M_reduced;
//   Number fscaling;
//   Number SurfaceTensionScaling;
//
//   //Number rho (const Number& x) {return rho_0;}
//
//
//
//
//
//   // Constructor takes parameter
//   Problem (const Number& dt_, const Dune::ParameterTree& ptree)
//            : t(0.0)
//   {
//     eps = ptree.get("problem.eps",(Number)0.1);
//     delta = ptree.get("problem.delta",(Number)0.1);
//     u_regularization = ptree.get("problem.URegularization",(Number)0.01);
//
//
//     fscaling = ptree.get("phsyics.fscaling",(Number)1);
//
//
//     dt=dt_;
//   }
//

//
//
//   Number f (const Number& u) const
//   {
//     if(u<0)
//       return fscaling*(-1.0+1000*u);
//     else
//       return fscaling*(u*u-1);
//   }
//
//   Number q (const Number& pf) const
//   {
//     if (pf<0 || pf>1)
//       return 0;
//     return pf*(1-pf);
//   }
//
//   Number q (const Number& pf1, const Number& pf2, const Number& pf3) const
//   {
//     if (pf1<0 || pf3<0)
//       return 0;
//     return pf1*pf3;
//   }
//
//   Number Diffusion(const Number& pf_c)
//   {
//     return D*fmax(1e-3,(pf_c-0.05)/0.95);
//   }
//
//   Number VelDissipation(const Number& pf_f)
//   {
//     return 10*(1-pf_f)*(1-pf_f);
//   }
//
//   //! Set time in instationary case
//   void setTime (Number t_)
//   {
//     t = t_;
//   }
//
// };
