
#pragma once
//#include "../myincludes.hh"
#include <dune/phasefield/base_potentials.hh>
#include <cmath>
#include <limits>
#include <dune/common/parametertreeparser.hh>
#include <iostream>


//______________________________________________________________________________________________________


template<typename Number>
class DoubleWell_poly
{
public:
  const Number TurningPointLeft  = 0.5 - 1/sqrt(12);
  const Number TurningPointRight = 0.5 + 1/sqrt(12);

  DoubleWell_poly(const Dune::ParameterTree & ptree)
  {
  }

  //Double Well
  Number eval(const Number& x) const
  {
    return x*x*(1-x)*(1-x);
  }

  // Double Well Derivative
  Number eval_d(const Number& x) const
  {
    return 2*x*(1-x)*(1-x) - 2*x*x*(1-x);
  }

  // Double Well Derivative(/Secant)
  Number eval_d (const Number& x, const Number& y) const
  {
    return (x+y)*(1 +x*x +y*y) -2*(x*x +x*y +y*y);
  }

  // Double Well Derivative of Convex Part
  Number eval_dConvex (const Number& x) const
  {
    if(x<TurningPointLeft)
      return eval_d(x)- eval_d(TurningPointLeft);
    if(x>TurningPointRight)
      return eval_d(x)-eval_d(TurningPointRight);
    return 0;
  }

  // Double Well Derivative of Convcarve Part
  Number eval_dConcarve (const Number& x) const
  {
    if(x<TurningPointLeft)
      return eval_d(TurningPointLeft);
    if(x>TurningPointRight)
      return eval_d(TurningPointRight);
    return eval_d(x);
  }

  // The function satisfying q=sqrt(2 W), inbetween 0 and 1
  Number eval_q(const Number& x) const
  {
    if(x<0 || x>1)
      return 0;
    else
      return sqrt(2)*x*(1-x);
  }
};

//______________________________________________________________________________________________________

template<typename Number>
class Limiter_Diverging
{
public:
  const Number delta;


  Limiter_Diverging(const Dune::ParameterTree & ptree) :
    delta(ptree.get("DoubleWellDelta",(Number)0.1))
  {
  }

  Number eval(const Number& x) const
  {
    if(x<=-delta)
    {
      return std::numeric_limits<Number>::max();
    }
    if(x>=0)
    {
      return 0;
    }
    return x*x/(delta+x);
  }

  Number eval_d(const Number& x) const
  {
    if(x<=-delta)
    {
      return std::numeric_limits<Number>::max();
    }
    if(x>=0)
    {
      return 0;
    }
    return x*(x+2*delta)/((x+delta)*(x+delta));
  }
};

//______________________________________________________________________________________________________

template<typename Number>
class Limiter_FakeDiverging
{
public:
  const Number delta; //Where the potential would be diverging
  const Number gamma; //Where we go into a linear function
  Number gammaeval;
  Number gammaeval_d;
  Limiter_FakeDiverging(const Dune::ParameterTree & ptree) :
    delta(ptree.get("DoubleWellDelta",(Number)0.1)),
    gamma(ptree.get("DoubleWellGamma",(Number)delta/20))
  {
    gammaeval=delta*gamma*gamma/(delta-gamma);
    gammaeval_d=-delta*gamma*(-gamma+2*delta)/((delta-gamma)*(delta-gamma));
  }

  Number eval(const Number& x) const
  {
    if(x<=-gamma)
    {
      std::cout << "Used FakeDiverging" <<std::endl;
      return gammaeval + (x-gamma)*gammaeval_d;
    }
    if(x>=0)
    {
      return 0;
    }
    return delta*x*x/(delta+x);
  }

  Number eval_d(const Number& x) const
  {
    if(x<=-gamma)
    {
      //std::cout << "Used FakeDiverging" <<std::endl;
      return gammaeval_d;
    }
    if(x>=0)
    {
      return 0;
    }
    return delta*x*(x+2*delta)/((x+delta)*(x+delta));
  }
};

//______________________________________________________________________________________________________

template<typename Number, typename DoubleWell, typename Limiter>
class DoubleWell_limited
{
public:
  DoubleWell wp;
  Limiter limiter;

  DoubleWell_limited(const Dune::ParameterTree & ptree) :
    wp(ptree),
    limiter(ptree)
  {
  }

  //Double Well
  Number eval(const Number& x) const
  {
    return wp.eval(x) + limiter.eval(x) + limiter.eval(1-x);
  }

  // Double Well Derivative
  Number eval_d(const Number& x) const
  {
    return wp.eval_d(x) + limiter.eval_d(x) - limiter.eval_d(1-x);
  }

  // Double Well Derivative(/Secant)
  Number eval_d (const Number& x, const Number& y) const
  {
    std::cout << "DoubleWell_limited secant not properly implemented yet" << std::endl;
    if(x==y)
    {
      return eval_d(x);
    }
    else
    {
      return (eval(x)-eval(y))/(x-y);
    }
  }

  // Double Well Derivative of Convex Part
  Number eval_dConvex (const Number& x) const
  {
    return wp.eval_dConvex(x) + limiter.eval_d(x) - limiter.eval_d(1-x);
  }

  // Double Well Derivative of Convcarve Part
  Number eval_dConcarve (const Number& x) const
  {
    return wp.eval_dConcarve(x);
  }

  // Just foward q, we do not care about the limiter for this
  Number eval_q (const Number& x) const
  {
    return wp.eval_q(x);
  }
};


//______________________________________________________________________________________________________

template<typename Number, typename DW>
class TripleWell
{
public:
  const Number Sigma1;
  const Number Sigma2;
  const Number Sigma3;
  const Number SigmaT;

  DW dw;

  TripleWell(const Dune::ParameterTree & param) :
    Sigma1(param.get("Sigma1",(Number)1)),
    Sigma2(param.get("Sigma2",(Number)1)),
    Sigma3(param.get("Sigma3",(Number)1)),
    SigmaT(1/(1/Sigma1 + 1/Sigma2 + 1/Sigma3)),
    dw(param)
  {
  }

  Number eval_dpf1 (const Number& pf1, const Number& pf2, const Number& pf3) const
  {
    return (Sigma1-SigmaT)*dw.eval_d(pf1)-SigmaT*dw.eval_d(pf2)-SigmaT*dw.eval_d(pf3);
  }

  Number eval_dpf2 (const Number& pf1, const Number& pf2, const Number& pf3) const
  {
    return (Sigma2-SigmaT)*dw.eval_d(pf2)-SigmaT*dw.eval_d(pf1)-SigmaT*dw.eval_d(pf3);
  }

  Number eval_dpf3 (const Number& pf1, const Number& pf2, const Number& pf3) const
  {
    return (Sigma3-SigmaT)*dw.eval_d(pf3)-SigmaT*dw.eval_d(pf1)-SigmaT*dw.eval_d(pf2);
  }

  Number eval_q  (const Number& pf1, const Number& pf2, const Number& pf3) const
  {
    if(abs(pf2-1)<1e-14)
      return 0;
    return dw.eval_q(pf1/(1-pf2))*(1-pf2)*(1-pf2);  //A function being q on the 13 interface and 0 on the others
  }
};
