#pragma once

#include<dune/common/parametertreeparser.hh>

#include<dune/pdelab/finiteelementmap/p0fem.hh>
#include <dune/pdelab/adaptivity/adaptivity.hh>
#include<dune/pdelab/gridoperator/gridoperator.hh>
#include<dune/pdelab/backend/istl.hh>

#include <dune/phasefield/base_parameters.hh>

//_________________________________________________________________________________________________________________________

template<typename Grid>
class chns_base
{
public:
  typedef typename Grid::LeafGridView GV;
  const int dim = GV::dimension;
  typedef double RF;


  chns_base(Grid& grid, Dune::ParameterTree& ptree)
  {
  }

};

class TimestepManager
{
public:
  double T;

  Params_Time& time;

  double dtMin;
  double dtMax;
  double dtIncreaseFactor;
  double dtDecreaseFactor;

  double lastsave;
  double saveintervall;
  double saveintervallstart;

  int verb;

  TimestepManager(const Dune::ParameterTree & param, Params_Time& time_, int verbosity) :
  time(time_)
  {
    verb = verbosity;
    T=param.get("tMax",(double)1.0);
    time.dt=param.get("dt",(double)0.1);

    dtMin = param.get("dtMin",(double)time.dt);
    dtMax = param.get("dtMax",(double)time.dt);
    dtIncreaseFactor = param.get("dtIncreaseFactor",(double)1.2);
    dtDecreaseFactor = param.get("dtDecreaseFactor",(double)0.5);

    saveintervall = param.get("saveintervall",(double)0.0);
    saveintervallstart = param.get("saveintervallstart",(double)0.0);

    time.t=param.get("tIni",(double)0.0);
    lastsave = time.t - saveintervall;
  }

  // RETURN VALUE: -1 -> ABORT, 0-> REDO TIMESTEP, 1-> OK
  template<typename NS_NEWTON, typename CH_NEWTON, typename VTKWRITER>
  int doTimestep(NS_NEWTON& ns_newton, CH_NEWTON& ch_newton, VTKWRITER& vtkwriter)
  {
    try
    {
      if (verb > 1)   std::cout << "= Begin newtonapply NS:" << std::endl;
      ns_newton.apply();

      if (verb > 1)   std::cout << "= Begin newtonapply CH:" << std::endl;
      ch_newton.apply();

    }
    catch(...)
    {
      if(time.dt <= dtMin)
      {
        std::cout << "===============================" << std::endl;
        if(dtMin==dtMax)
          std::cout << "Problem with Newton Solver, aborting :(" << std::endl;
        else
          std::cout << "Timestep too small, aborting :(" << std::endl;
        std::cout << "===============================" << std::endl;
        return -1;
      }
      time.dt = std::max(time.dt*dtDecreaseFactor,dtMin);
      if (verb > 0) std::cout << "Reducing Timestep to "<< time.dt << std::endl;

      return 0;
    }

    time.t += time.dt;
    if(time.dt < dtMax)
    {
      time.dt = std::min(time.dt*dtIncreaseFactor, dtMax);
      if (verb > 0)   std::cout << "New Timestep "<< time.dt << std::endl;
    }

    if( time.t < saveintervallstart || time.t>lastsave+saveintervall )
    {
        lastsave = time.t;
        vtkwriter.write(time.t,Dune::VTK::ascii);
    }
    return 1;
  }


  bool isFinished()
  {
    return (time.t>=T);
  }
};

//_________________________________________________________________________________________________________________________

template<typename GFS, typename NS_BDRY_FN, typename CH_BDRY_FN, typename NS_V, typename CH_V>
void interpolateToBdry(const GFS& gfs, const NS_BDRY_FN& nsBdryFn, const CH_BDRY_FN& chBdryFn, NS_V& nsV, CH_V& chV)
{
  NS_V nsInterp(gfs.ns);
  CH_V chInterp(gfs.ch);
  Dune::PDELab::interpolate(nsBdryFn, gfs.ns, nsInterp);
  Dune::PDELab::interpolate(chBdryFn, gfs.ch, chInterp);
  Dune::PDELab::copy_constrained_dofs(gfs.nsCC, nsInterp, nsV);
  Dune::PDELab::copy_constrained_dofs(gfs.chCC, chInterp, chV);
}

//_________________________________________________________________________________________________________________________

template<typename GV, typename RF, typename EST_LOP, typename CH_GFS>
class RefinementIndicator
{
public:
  //Estimator GFS
  typedef Dune::PDELab::P0LocalFiniteElementMap<RF,RF,GV::dimension> P0FEM;
  P0FEM p0fem;
  typedef Dune::PDELab::NoConstraints NCON;
  typedef Dune::PDELab::GridFunctionSpace<GV,P0FEM,NCON,Dune::PDELab::ISTL::VectorBackend<> > P0GFS;
  P0GFS p0Gfs;
  //Estimator vector
  using Z0 = Dune::PDELab::Backend::Vector<P0GFS,RF>;
  Z0 z0;

  typedef Dune::PDELab::DiscreteGridFunction<P0GFS,Z0> Z0DGF;
  Z0DGF z0dgf;

  //Estimator Operator
  EST_LOP estlop;
  typedef Dune::PDELab::EmptyTransformation NCC;
  typedef Dune::PDELab::ISTL::BCRSMatrixBackend<> MBE;
  typedef Dune::PDELab::GridOperator<CH_GFS,P0GFS,EST_LOP,MBE,RF,RF,RF,NCC,NCC> ESTGO;
  ESTGO estgo;

  double eta_refine;
  double eta_coarsen;
  int Rmaxlevel;
  int Rminlevel;

  RefinementIndicator(const GV& gv, const CH_GFS chGfs, const Dune::ParameterTree & param) :
    p0fem(Dune::GeometryTypes::simplex(GV::dimension)),
    p0Gfs(gv,p0fem),
    z0(p0Gfs,0.0),
    z0dgf(p0Gfs,z0),
    estlop(),
    estgo(chGfs,p0Gfs,estlop,MBE(1))
  {
    eta_refine = param.get("etaRefine",(double)0.15);
    eta_coarsen = param.get("etaCoarsen",(double)0.05);
    Rmaxlevel = param.get("maxRefineLevel",(double)2);
    Rminlevel = param.get("minRefineLevel",(double)0);
  }

  template<typename CH_V>
  const Z0& calculateIndicator(const GV& gv, const CH_V& chV, bool writeOutput)
  {
    estgo.residual(chV,z0);
    if(writeOutput)
    {
      auto vtkwriter = std::make_shared<Dune::SubsamplingVTKWriter<GV> >(gv,0);
      vtkwriter->addVertexData(std::make_shared<Dune::PDELab::VTKGridFunctionAdapter<Z0DGF> >(z0dgf,"indicator"));
      vtkwriter->write("RefinementIndicator",Dune::VTK::ascii);
    }
    return z0;
  }

  template<typename Grid>
  void markGrid(Grid &grid, int verbosity=0)
  {
    if(verbosity>0) std::cout<<"mark elements"<<std::endl;
    Dune::PDELab::mark_grid(grid,z0,eta_refine,eta_coarsen, Rminlevel,Rmaxlevel,verbosity,-1);
  }
};
