
#pragma once
#include <cmath>
#include <dune/pdelab/common/function.hh>
#include <dune/phasefield/base_parameters.hh>


template<typename GV, typename RF, std::size_t dim_range, typename Params>
class ZeroInitialVector :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim_range>,ZeroInitialVector<GV,RF,dim_range, Params> >
{
public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim_range> Traits;

  ZeroInitialVector(const GV & gv, const Params & params) :
    Dune::PDELab::AnalyticGridFunctionBase<Traits, ZeroInitialVector>(gv) {}

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y=0;
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params>
class ZeroInitial
  : public ZeroInitialVector<GV,RF,1,Params>
{
public:
  ZeroInitial(const GV & gv, const Params & params) : ZeroInitialVector<GV,RF,1,Params>(gv,params) {}
};

//______________________________________________________________________________________________________________________

#define CONST_INITIAL_U 0

template<typename GV, typename RF, typename Params, const int type>
class ConstInitial :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, ConstInitial<GV,RF,Params, type> >
{
public:
  RF myval;
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  ConstInitial(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,ConstInitial<GV,RF,Params,type> > (gv)
     {
        if(type == CONST_INITIAL_U)
           myval = params.initial.ptree.get("uConst",(RF)1);
        else
           myval = 0;
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = myval;
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params>
class PhiInitial_RayleighTaylor :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_RayleighTaylor<GV,RF,Params> >
{
  RF eps;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_RayleighTaylor(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_RayleighTaylor<GV,RF,Params> > (gv)
     {
       eps = params.eps;
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = 0.5*(1-tanh(1/eps/sqrt(2)*(x[1]-0.75-0.01*cos(x[0]*2*3.1415))));
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params, const int Phase>
class PhiInitial_3pContactPoint :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_3pContactPoint<GV,RF,Params,Phase> >
{
  RF eps;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_3pContactPoint(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_3pContactPoint<GV,RF,Params,Phase> > (gv)
     {
       eps = params.eps;
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = 0.0;

    RF d1 = x[0]-0.5;
    RF yx = 0.5*(1+tanh(1/eps/sqrt(2)*d1));

    RF d2 = x[1]-0.75;
    if(Phase == 0)
      d2 = -d2;
    y[0]=0.5*(1+tanh(1/eps/sqrt(2)*d2))*yx;
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params, const int Phase>
class PhiInitial_3pFlowTest :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_3pFlowTest<GV,RF,Params,Phase> >
{
  RF eps;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_3pFlowTest(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_3pFlowTest<GV,RF,Params,Phase> > (gv)
     {
       eps = params.eps;
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = 0.0;

    RF d1 = x[0]-0.3;


    RF d2 = x[1]-0.25;
    RF yy = 0.5*(1+tanh(1/eps/sqrt(2)*d2));
    if(Phase == 0)
      d1 = -d1;
    y=0.5*(1+tanh(1/eps/sqrt(2)*d1))*yy;
  }
};
//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params>
class PhiInitial_Sphere :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_Sphere<GV,RF,Params> >
{
  RF eps;
  RF radius;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_Sphere(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_Sphere<GV,RF,Params> > (gv)
     {
       eps = params.eps;
       radius = params.initial.ptree.get("Radius",0.2);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    RF dist = sqrt( (x[0]-0.5)*(x[0]-0.5) + (x[1]-0.5)*(x[1]-0.5) );
    y[0] = 0.5*(1+tanh(1/eps/sqrt(2)*(dist-radius)));
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params>
class PhiInitial_Rectangle :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_Rectangle<GV,RF,Params> >
{
  RF xcenter;
  RF ycenter;
  RF eps;
  RF radius;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_Rectangle(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_Rectangle<GV,RF,Params> > (gv)
     {
       eps = params.eps;
       radius = params.initial.ptree.get("Radius",0.2);
       xcenter = params.initial.ptree.get("xCenter",0.5);
       ycenter = params.initial.ptree.get("yCenter",0.5);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    RF dist1 = std::abs(x[0]-xcenter);
    RF dist2 = std::abs(x[1]-ycenter);
    RF m1 = 0.5*(1+tanh(1/eps/sqrt(2)*(dist1-radius)));
    RF m2 = 0.5*(1+tanh(1/eps/sqrt(2)*(dist2-radius)));
    y[0] = 1-(1-m1)*(1-m2);
  }
};

//______________________________________________________________________________________________________________________

template<typename GV, typename RF, typename Params>
class PhiInitial_2Precip :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_2Precip<GV,RF,Params> >
{
  RF eps;
  RF x1center;
  RF y1center;
  RF radius1;
  RF x2center;
  RF y2center;
  RF radius2;
  RF x3center;
  RF y3center;
  RF radius3;


public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_2Precip(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_2Precip<GV,RF,Params> > (gv)
     {
       eps = params.eps;
       radius1 = params.initial.ptree.get("Radius1",0.2);
       x1center = params.initial.ptree.get("x1Center",0.5);
       y1center = params.initial.ptree.get("y1Center",0.5);
       radius2 = params.initial.ptree.get("Radius2",0.2);
       x2center = params.initial.ptree.get("x2Center",0.5);
       y2center = params.initial.ptree.get("y2Center",0.5);
       radius3 = params.initial.ptree.get("Radius3",0.2);
       x3center = params.initial.ptree.get("x3Center",0.5);
       y3center = params.initial.ptree.get("y3Center",0.5);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    RF dist1 = std::sqrt( (x[0]-x1center)*(x[0]-x1center) + (x[1]-y1center)*(x[1]-y1center) );
    RF y1 = 0.5*(1+tanh(1/eps/sqrt(2)*(dist1-radius1)));

    RF dist2 = std::sqrt( (x[0]-x2center)*(x[0]-x2center) + (x[1]-y2center)*(x[1]-y2center) );
    RF y2 = 0.5*(1+tanh(1/eps/sqrt(2)*(dist2-radius2)));

    RF dist3 = std::sqrt( (x[0]-x3center)*(x[0]-x3center) + (x[1]-y3center)*(x[1]-y3center) );
    RF y3 = 0.5*(1+tanh(1/eps/sqrt(2)*(dist3-radius3)));

    y[0] = y1*y2*y3;
  }
};

template<typename GV, typename RF, typename Params>
class PhiInitial_SummerSchoolFull :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_SummerSchoolFull<GV,RF,Params> >
{
  RF eps;
  RF radius;
  std::vector<RF> xValues;
  std::vector<RF> yValues;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_SummerSchoolFull(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_SummerSchoolFull<GV,RF,Params> > (gv)
     {
       eps = params.eps;
       radius = params.initial.ptree.get("Radius",0.2);

       xValues.push_back(0.560); yValues.push_back(0.070);
       xValues.push_back(0.831); yValues.push_back(0.029);
       xValues.push_back(1.353); yValues.push_back(0.008);
       xValues.push_back(1.298); yValues.push_back(0.105);
       xValues.push_back(1.387); yValues.push_back(0.108);
       xValues.push_back(1.959); yValues.push_back(0.014);
       xValues.push_back(2.397); yValues.push_back(-0.001);
       xValues.push_back(3.014); yValues.push_back(0.074);
       xValues.push_back(3.703); yValues.push_back(0.244);
       xValues.push_back(3.808); yValues.push_back(-0.015);
       xValues.push_back(4.190); yValues.push_back(0.071);
       xValues.push_back(4.816); yValues.push_back(0.026);

     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    RF dist;
    RF temp;
    RF yc = 1;
    for(int i=0; i<xValues.size(); i++)
    {
      dist = std::sqrt( (x[0]-xValues[i])*(x[0]-xValues[i]) + (x[1]-yValues[i])*(x[1]-yValues[i]) );
      temp = 0.5*(1+tanh(1/eps/sqrt(2)*(dist-radius)));
      yc = yc*temp;
    }
    y[0] = yc;
  }
};

//______________________________________________________________________________________________________________________

#include <dune/phasefield/boundary_fn.hh>

template<typename GV, typename RF, typename Params, int dim>
class VelocityInitial_Inflow :
  public Dune::PDELab::AnalyticGridFunctionBase<
  Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim>,
  VelocityInitial_Inflow<GV,RF,Params,dim> >
{
  const Params_Time& time;
  RF meanflow;


public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim> Traits;

  typedef typename Traits::DomainType DomainType;
  typedef typename Traits::RangeType RangeType;

  VelocityInitial_Inflow(const GV & gv, const Params & params) :
    Dune::PDELab::AnalyticGridFunctionBase<Traits,VelocityInitial_Inflow<GV,RF,Params,dim> > (gv)
    , time(params.time)
  {
    meanflow = params.initial.ptree.get("MeanInflow",(RF)0.1);
  }

  template <typename T>
  void setTime(T t){
    time = t;
  }

  inline void evaluateGlobal(const DomainType & x, RangeType & y) const
  {
    y = 0.0;
    if(x[0]<GRID_EPS) //Inflow boundary
    {
      RF factor = meanflow * 6/( (GRID_TOP-GRID_BOTTOM)*(GRID_TOP-GRID_BOTTOM) );
      y[0] = std::max((RF)0, factor * (x[1]-GRID_BOTTOM)*(GRID_TOP-x[1]));
    }
  }
};


template<typename GV, typename RF, typename Params>
class UInitial_Inflow :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, UInitial_Inflow<GV,RF,Params> >
{
  RF uConst;
  RF uInflow;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  UInitial_Inflow(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,UInitial_Inflow<GV,RF,Params> > (gv)
     {
       uConst = params.initial.ptree.get("uConst",(RF)1.0);
       uInflow = params.initial.ptree.get("uInflow",(RF)1.5);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = uConst;
    if(x[0]<1e-7)
    {
      y=uInflow;
    }
  }
};

template<typename GV, typename RF, typename Params>
class PInitial_Inflow :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PInitial_Inflow<GV,RF,Params> >
{

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PInitial_Inflow(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PInitial_Inflow<GV,RF,Params> > (gv)
     {
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = 0;
    if(x[0]<1e-7)
    {
      y=20;
    }
  }
};
