#pragma once

// C includes
#include<sys/stat.h>
// C++ includes
#include<iostream>
// Dune includes
#include<dune/grid/io/file/vtk.hh>
#include<dune/pdelab/newton/newton.hh>

#include <dune/phasefield/localoperator/fem_ffs_chns_r_cahnhilliard.hh>
#include <dune/phasefield/localoperator/fem_ffs_chns_navierstokes.hh>
#include <dune/phasefield/localoperator/pf_3p_diff_estimator.hh>

#include <dune/phasefield/utility/ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES.hh>

#include <dune/phasefield/chns_base.hh>
#include <dune/phasefield/ffs_chns_r.hh>
#include <dune/phasefield/initial_fn.hh>
#include <dune/phasefield/debug/vector_debug.hh>



//_________________________________________________________________________________________________________________________

template<typename Grid, typename ES, typename V_FEM, typename P_FEM, typename PHI_FEM>
void driver_3p (Grid& grid, const ES& es, const V_FEM& vFem, const P_FEM pFem, const PHI_FEM& phiFem, Dune::ParameterTree& ptree, Dune::MPIHelper& helper)
{

  typedef typename Grid::LeafGridView GV;
  //std::bitset<2> periodic_directions(std::string("01"));
  //typedef Dune::XT::Grid::PeriodicGridView<LGV> GV;
  GV gv = grid.leafGridView();
  const int dim = GV::dimension;
  typedef double RF;


  typedef Params_ffs_r< RF,
                        TripleWell<RF,DoubleWell_limited< RF, DoubleWell_poly<RF>, Limiter_FakeDiverging<RF> > >,
                        Mobility_Const<RF>,
                        Reaction_None<RF>,
                        VelDissipation_Quadratic<RF> > Parameters;
  Parameters param(ptree);
  TimestepManager timestepManager(ptree.sub("Time"),param.time,2);


//_________________________________________________________________________________________________________________________
// Make grid function spaces


  typedef GFS_ffs_chns_r<RF, ES, V_FEM, P_FEM, PHI_FEM, PHI_FEM> GFS;
  GFS gfs(es,vFem,pFem,phiFem,phiFem);
  typedef typename GFS::NS NS_GFS;
  typedef typename GFS::CH CH_GFS;


  //_________________________________________________________________________________________________________________________
  //Constraints

  typedef Bdry_ffs_chns_r_compositeV_2d<InflowOutflow_v,            // v_x
              Dune::PDELab::DirichletConstraintsParameters,         // v_y
              InflowOutflow_p,                                      // p
              InflowOutflow_phi,                                    // phi1
              Dune::PDELab::NoDirichletConstraintsParameters,       // mu1
              InflowOutflow_phi,                                    // phi2
              Dune::PDELab::NoDirichletConstraintsParameters,       // mu2
              InflowOutflow_u>                                      // u
              Bdry;
  Bdry bdry;
  gfs.updateConstraintsContainer(bdry);

  //_________________________________________________________________________________________________________________________
  //Coefficient vectors

  using CH_V = Dune::PDELab::Backend::Vector<CH_GFS, RF>;
  using NS_V = Dune::PDELab::Backend::Vector<NS_GFS, RF>;
  CH_V chV(gfs.ch);
  NS_V nsV(gfs.ns);
  CH_V chVOld = chV;
  NS_V nsVOld = nsV;

  //_________________________________________________________________________________________________________________________
  //Initial Values

    typedef Ini_ffs_chns_r< VelocityInitial_Inflow<GV,RF,Parameters,dim>,
                      ZeroInitial<GV,RF,Parameters>,
                      PhiInitial_3pFlowTest<GV,RF,Parameters,0>,
                      ZeroInitial<GV,RF,Parameters>,
                      PhiInitial_3pFlowTest<GV,RF,Parameters,1>,
                      ZeroInitial<GV,RF,Parameters>,
                      ZeroInitial<GV,RF,Parameters> >  Initial;
    Initial initial(gv,param);
    Dune::PDELab::interpolate(initial.iniCH , gfs.ch, chV);
    Dune::PDELab::interpolate(initial.iniNS , gfs.ns, nsV);

  //_________________________________________________________________________________________________________________________
  // LocalOperators

  typedef FEM_FFS_CHNS_NavierStokes<Parameters, CH_GFS, CH_V, NS_GFS, NS_V> NS_LOP;
  NS_LOP nsLop(param, gfs.ch, chV, chVOld, gfs.ns, nsVOld);

  typedef FEM_FFS_CHNS_R_CahnHilliard<Parameters, CH_GFS, CH_V, NS_GFS, NS_V> CH_LOP;
  CH_LOP chLop(param, gfs.ch, chVOld, gfs.ns, nsV);

  typedef Dune::PDELab::ISTL::BCRSMatrixBackend<> MBE;
  MBE mbe(10);// Maximal number of nonzeroes per row can be cross-checked by patternStatistics().


  //_________________________________________________________________________________________________________________________
  // Construct discrete Grid Functions
  typedef DGF_ffs_chns_r<GFS, NS_V, CH_V> DGF;
  DGF dgf(gfs, nsV, chV);

  //_________________________________________________________________________________________________________________________
  // prepare VTK writer
  std::string filename=ptree.get("output.filename","output");
  struct stat st;
  if( stat( filename.c_str(), &st ) != 0 )
  {
    int stat = 0;
    stat = mkdir(filename.c_str(),S_IRWXU|S_IRWXG|S_IRWXO);
    if( stat != 0 && stat != -1)
      std::cout << "Error: Cannot create directory "  << filename << std::endl;
  }
  auto stationaryvtkwriter = std::make_shared<Dune::SubsamplingVTKWriter<GV> >(gv,0);
  Dune::VTKSequenceWriter<GV> vtkwriter(stationaryvtkwriter,filename.c_str(),filename.c_str(),"");
  dgf.addToVTKWriter(vtkwriter);

  vtkwriter.write(param.time.t,Dune::VTK::ascii);

//_________________________________________________________________________________________________________________________
  printVector(chV,10,"chV");
  printVector(nsV,10,"nsV");


  //____________________________________________________________________________________________________________________
  //int i_it=0;
  chVOld = chV;
  nsVOld = nsV;
  while (!timestepManager.isFinished())
    {

      // Grid Operators
      typedef Dune::PDELab::GridOperator<NS_GFS,NS_GFS,NS_LOP,MBE,RF,RF,RF,typename GFS::NS_CC,typename GFS::NS_CC> NS_GO;
      NS_GO nsGo(gfs.ns,gfs.nsCC,gfs.ns,gfs.nsCC,nsLop,mbe);

      typedef Dune::PDELab::GridOperator<CH_GFS,CH_GFS,CH_LOP,MBE,RF,RF,RF,typename GFS::CH_CC,typename GFS::CH_CC> CH_GO;
      CH_GO chGo(gfs.ch,gfs.chCC,gfs.ch,gfs.chCC,chLop,mbe);

      // Linear solver
      typedef ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES
                  <CH_GO, Dune::SeqILU0, Dune::RestartedGMResSolver> CH_LS; CH_LS ls_ch(chGo,500,100,3,1);

      typedef ISTLBackend_NOVLP_BASE_PREC_FOR_GMRES
                  <NS_GO, Dune::SeqILU0, Dune::RestartedGMResSolver> NS_LS; NS_LS ls_ns(nsGo,1000,200,3,1);

      // Nonlinear solver
      typedef Dune::PDELab::Newton<NS_GO,NS_LS,NS_V> PDESolver1;
      PDESolver1 ns_newton(nsGo,nsV,ls_ns);
      ns_newton.setParameters(ptree.sub("NS_Newton"));

      typedef Dune::PDELab::Newton<CH_GO,CH_LS,CH_V> PDESolver2;
      PDESolver2 ch_newton(chGo,chV,ls_ch);
      ch_newton.setParameters(ptree.sub("Phasefield_Newton"));

      //_________________________________________________________________________________________________________________________
      // do time step


      interpolateToBdry(gfs,initial.iniNS,initial.iniCH, nsV, chV);

      int result = timestepManager.doTimestep(ns_newton,ch_newton,vtkwriter);
      if(result <0)
      {
        return;
      }
      else if(result == 0)
      {
        nsV = nsVOld;
        chV = chVOld;
        continue;
      }

      chVOld = chV;
      nsVOld = nsV;

      //_________________________________________________________________________________________________________________________

      // mark elements for refinement
      typedef RefinementIndicator<GV,RF,PF_3P_DIFF_ESTIMATOR<CH_GFS>,CH_GFS> RIND;
      RIND rInd(gv,gfs.ch,ptree.sub("Refinement"));
      rInd.calculateIndicator(gv,chV,false);
      rInd.markGrid(grid,2);

      printVector(chV,3,"chV");
      printVector(nsV,3,"nsV");

      // do refinement
      std::cout << "adapt grid and solution" << std::endl;
      auto chTransfer = transferSolutions(gfs.ch,4,chV,chVOld);
      auto nsTransfer = transferSolutions(gfs.ns,4,nsV,nsVOld);
      Dune::PDELab::adaptGrid(grid, chTransfer,nsTransfer);
      std::cout << "adapt grid and solution ended" << std::endl;

      printVector(chV,3,"chV");
      printVector(nsV,3,"nsV");

      //vtkwriter.write(timestepManager.t+1e-6,Dune::VTK::ascii);

      // recompute constraints
      gfs.updateConstraintsContainer(bdry);
      interpolateToBdry(gfs,initial.iniNS,initial.iniCH,nsV,chV);
      interpolateToBdry(gfs,initial.iniNS,initial.iniCH,nsVOld,chVOld);


      //vtkwriter.write(timestepManager.t+2e-6,Dune::VTK::ascii);
    }
}
